/**********************************************************************
 ***** Pierre Hyvernat, cours info505 "Algorithmes de graphes"    *****
 *****                                                            *****
 ***** Il s'agit des définitions de quelques fonctions pour le TP *****
 ***** en salles machines...                                      *****
 ***** Les commentaire pertinents se trouvent dans le fichier     *****
 ***** "graphes.h".                                               *****
 *********************************************************************/



#include "graphes.h"


void ajoute (int o, int p, int b, ListeAdj *L) {
 ListeAdj l = malloc(sizeof(struct __arc)) ;
 l->origine = o ;
 l->poids = p ;
 l->but = b ;
 l->suivant = *L ;
 l->precedent = NILL ;
 if (*L!=NULL) (*L)->precedent = l ;
 *L = l ;
}



Tile tileVide() {
 Tile T = malloc(sizeof(struct __Tile)) ;
 T->premier = NULL ;
 T->dernier = NULL ;
 return(T) ;
}


void entile(int o, int p, int b, Tile T) {

 struct __arc *N = malloc(sizeof(struct __arc)) ;
 N->origine = o ;
 N->poids = p ;
 N->but = b ;
 N->precedent = NULL ;
 N->suivant = T->premier ;

 if (T->dernier == NULL) { T->dernier = N ;}
 else { T->premier->precedent = N ;}

 T->premier = N ;
}


void depile(Tile T, int *o, int *p, int *b) {
 if (o!=NULL) *o = T->premier->origine ;
 if (p!=NULL) *p = T->premier->poids ;
 if (b!=NULL) *b = T->premier->but ;

 if (T->premier == T->dernier) { /* la tile a un seul élément */
  free(T->premier) ;
  T->premier = NULL ;
  T->dernier = NULL ;
 }
 else {
  T->premier = T->premier->suivant ;
  free(T->premier->precedent) ;
  T->premier->precedent = NULL ;
 }
}


void defile(Tile T, int *o, int *p, int *b) {
 if (o!=NULL) *o = T->dernier->origine ;
 if (p!=NULL) *p = T->dernier->poids ;
 if (b!=NULL) *b = T->dernier->but ;

 if (T->premier == T->dernier) { /* la tile a un seul élément */
  free(T->premier) ;
  T->premier = NULL ;
  T->dernier = NULL ;
 }
 else {
  T->dernier= T->dernier->precedent ;
  free(T->dernier->suivant) ;
  T->dernier->suivant = NULL ;
 }
}

int tileEstVide(Tile T) {
  return(T->dernier == NULL) ;
}





GrapheListe grapheListeVide (int n) {
 int i;
 GrapheListe G = malloc(sizeof(struct __GrapheListe));
 G->nb_sommets = n ;
 G->Adj = malloc(n*sizeof(ListeAdj*)) ;
 for(i=0;i<n;i++) G->Adj[i]=NILL;
 return(G) ;
}


GrapheMatr grapheMatrVide (int n) {
  GrapheMatr G=malloc(sizeof(struct __GrapheMatr)) ;
  int i ;
  G->nb_sommets = n ;
  G->Matr = malloc(n*sizeof(int*)) ;
  for(i=0;i<n;i++) G->Matr[i] = calloc(n,sizeof(int)) ;
  return(G) ;
}


GrapheListe lireGrapheListe(const char *fichier) {
 FILE *flot;
 GrapheListe G ;
 int n,b,p,s ;
 if ((flot = fopen(fichier, "r")) == NULL) {
  fprintf(stderr,"\nErreur: impossible d'ouvrir le fichier %s\n", fichier);
  return(NULL);
 }
 if (!fscanf(flot, "%i", &n)) {
  fprintf(stderr,"\nErreur: je n'arrive pas à lire les nombre de sommets dans %s\n", fichier);
  return(NULL);
 }
 G = grapheListeVide(n) ;
 while (fscanf(flot, "%i %i %i", &s, &p, &b) == 3) {
  if (p!=0 && s>=0 && s<n && b>=0 && b<n) ajoute(s,p,b,&(G->Adj[s])) ;
 }
 return(G) ;
}


GrapheListe grapheListeAleatoire(int n, int m, int p, int o) {
 struct timeval temps ;
 int i, q, a, s, t;
 int N=n*n ;
 GrapheListe G = grapheListeVide(n) ;
 gettimeofday(&temps, NULL) ;
 srand(temps.tv_usec) ;
 for(i=0;i<m;i++) {
  q=1+(int) ((float)p*rand()/(RAND_MAX+1.0));
  a=(int) ((float)N*rand()/(RAND_MAX+1.0));
  s = a%n ;
  t = a/n ;
  ajoute(t,q,s,&(G->Adj[t])) ;
  if (o!=0) ajoute(s,q,t,&(G->Adj[s]));
 }
 return(G) ;
}



/* fonction temproraire pour trier les voisins dans les listes d'adjacence */
static void insere(int o, int p, int b, ListeAdj *l) {
 int oo, pp, bb ;
 ListeAdj ll = *l;
 if (ll==NULL) { ajoute(o,p,b,&ll) ; }
 else {
  pp = ll->poids ;
  bb = ll->but ;
  oo = ll->origine ;
  if (p<pp) { ajoute(o,p,b,&ll) ; }
  else {
   ll = ll->suivant ;
   insere(o,p,b,&ll) ;
   ajoute(oo,pp,bb,&ll) ;
  }
 }
 *l = ll ;
}


GrapheListe grilleRectPleine(int largeur, int hauteur) {

 int h,l,s,p ;
 GrapheListe G = grapheListeVide(largeur*hauteur) ;

 struct timeval temps ;
 gettimeofday(&temps, NULL) ;
 srand(temps.tv_usec) ;

 for (h=0;h<hauteur;h++)
 for (l=0;l<largeur;l++) {
  s=l+h*largeur ;
  if (h!=0) {
   p=rand() ;
   insere(s,p,s-largeur,&(G->Adj[s])) ;
   insere(s-largeur,p,s,&(G->Adj[s-largeur])) ;
  }
  if (l!=0) {
   p=rand() ;
   insere(s,p,s-1,&(G->Adj[s])) ;
   insere(s-1,p,s,&(G->Adj[s-1])) ;
  }
 }
 return(G) ;
}


/* fonction locale qui vérifie si un sommet u a un voisin v dans le
graphe orienté G */
static int voisin_oriente(int u, int v, GrapheListe G) {
 int r=0 ;
 ListeAdj L=G->Adj[u] ;
 while (L!=NULL) {
  if (L->but == v) { r = 1 ; break ; }
  else { L = L->suivant ; }
 }
 return(r) ;
}

/* fonction locale qui vérifie si un sommet u a un voisin v dans le
graphe non-orienté G */
static int voisin(int u, int v, GrapheListe G) {
  int n = G->nb_sommets;
  return (0 <= u && u < n && 0 <= v && v<n) && (voisin_oriente(u,v,G) || voisin_oriente(v,u,G));
}

void afficheGrilleRect(GrapheListe G, int largeur, int hauteur, char SS[]) {
 int h,l,s ;
 char *S;
 if (SS==NULL) {
   S = malloc(largeur*hauteur*sizeof(char));
   for (s=0;s<largeur*hauteur;s++) S[s]=' ';
 }
 else S = SS;


 for(l=0;l<largeur;l++) { printf("+--") ;} printf("+\n") ;

 for(h=0;h<hauteur;h++) {
  printf("|") ;
  for(l=0;l<largeur;l++) {
   s = h*largeur+l ;
   printf(" %c",S[s]);
   //la case à droite de s est la case s+1
   if (voisin(s,s+1,G))
    printf(" ") ;
   else
    printf("|") ;
  }
  printf("\n") ;

  printf("+") ;
  for(l=0;l<largeur;l++) {
   s = h*largeur+l ;
   //la case en dessous de s est la case s+largeur
   if (voisin(s,s+largeur,G))
    printf("  +") ;
   else
    printf("--+") ;
  }
  printf("\n") ;
 }
}









GrapheListe grilleHexPleine(int largeur, int hauteur) {

 largeur = 2 * largeur;

 int h,l,s,p,t ;
 GrapheListe G = grapheListeVide(largeur*hauteur) ;

 struct timeval temps ;
 gettimeofday(&temps, NULL) ;
 srand(temps.tv_usec) ;

 for (h=0;h<hauteur;h++)
 for (l=0;l<largeur;l++) {
  s=h+(l*hauteur) ;
  t = s-1;
  if (h!=0) {  /* il y a une case au dessus */
   p=rand() ;
   insere(s,p,t,&(G->Adj[s])) ;
   insere(t,p,s,&(G->Adj[t])) ;
  }
  if (l!=0 && (l%2==1 || h!=0)) {  /* il y a une case au dessus à gauche */
   p=rand() ;
   t = l%2==0 ? s-hauteur-1 : s-hauteur ;
   insere(s,p,t,&(G->Adj[s])) ;
   insere(t,p,s,&(G->Adj[t])) ;
  }
  if (l!=largeur-1 && (l%2==1 || h!=0)) {  /* il y a une case au dessus à droite */
   p=rand() ;
   t = l%2==0 ? s+hauteur-1 : s+hauteur ;
   insere(s,p,t,&(G->Adj[s])) ;
   insere(t,p,s,&(G->Adj[t])) ;
  }
 }
 return(G) ;
}




void afficheGrilleHex(GrapheListe G, int largeur, int hauteur, char SS[]) {
 int h,l ;
 int s,u,v ;

 int n=G->nb_sommets;
 char *S;

 if (SS==NULL) {
   S = malloc(n*sizeof(char));
   int i;
   for (i=0;i<n;i++) {
     S[i] = ' ';
   }
 }
 else S=SS;

 /* premiere ligne */
 printf(" ");
 for(l=0;l<largeur;l++) printf("__    ") ;
 printf("\n");



 /* pour chaque numero de ligne : */
 for(h=0;h<hauteur;h++) {


   /* ligne du haut */
   printf("/");
   for(l=0;l<largeur;l++) {
     // on est en train de regarder la case (2l,h), ie la case n° 2l*hauteur +
     // h
     s = 2*l*hauteur + h;
     // il faut regarder les voisins bas de la case (2l+1,h-1), c à d les
     // voisins du bas de la case n° h-1+hauteur*(2l+1)
     u = hauteur * (2*l+1) + h - 1 ;
     printf(" %c",S[s]);
     // le voisin en bas à gauche de cette case est la case (2l,h), ie
     // 2l*hauteur + h
     v = hauteur * 2 * l + h;
     if (voisin(u,v,G)) printf(" ");
     else printf("\\");
     // le voisin en bas de cette case est la case (2l+1,h), ie
     // (2l+1)*hauteur + h
     v = hauteur * (2*l+1) + h;
     if (voisin(u,v,G)) printf("  ");
     else printf("__");
     // le voisin en bas à droite de cette case est la case (2l+2,h), ie
     // (2l+2)*hauteur + h
     v = hauteur * (2*l+2) + h;
     if (voisin(u,v,G) || (l==largeur-1 && h==0)) printf(" ");
     else printf("/");
   }
   printf("\n");


   /* ligne du bas */
   for(l=0;l<largeur;l++) {
     // on est en train de regarder la case (2l+1,h), ie la case n° (2l+1)*hauteur +
     // h
     s = (2*l+1)*hauteur + h;
     // il faut regarder les voisins bas de la case (2l,h), ie la case n°
     // 2l*hauteur + h
     u = 2*l*hauteur + h;
     // le voisin en bas à gauche de cette case est la case (2l-1,h), ie la
     // case (2l-1)hauteur + h
     v = (2*l-1) * hauteur + h;
     if (voisin(u,v,G)) printf(" ");
     else printf("\\");
     // le voisin en bas de cette case est la case (2l,h+1), ie la
     // case 2l*hauteur + h + 1
     v = 2*l*hauteur + h + 1;
     if (voisin(u,v,G)) printf("  ");
     else printf("__");
     // le voisin en bas à droite de cette case est la case (2l+1,h), ie la
     // case (2l+1)hauteur + h
     v = (2*l+1)*hauteur + h;
     if (voisin(u,v,G)) printf(" ");
     else printf("/");
     printf(" %c",S[s]);
   }
   printf("\\");


   printf("\n");
 }

 /* derniere ligne */
 printf(" ");
 for(l=0;l<largeur;l++) printf("  \\__/") ;
 printf("\n");


}





static int cmp__arc(const void *a1, const void *a2) {
  return ( ((struct __arc *)a1)->poids - ((struct __arc *)a2)->poids ) ;
}

GrapheListe kruskal (GrapheListe G) {
 int n=G->nb_sommets ;
 int m=0 ;
 int nb_aretes ;
 GrapheListe T = grapheListeVide(n) ;
 struct __arc *L ;  /* L est un tableau de "struct __arc" */
 //Tile L=tileVide() ;
 int *CC = malloc(n*sizeof(int)) ;
 GrapheListe CCL = grapheListeVide(n) ;
 int *taille = malloc(n*sizeof(int)) ;

 int x,y,t,i,j,c ;
 ListeAdj l ;
 struct __arc *TMP ;

 /* on compte le nombre d'arcs */
 nb_aretes=0;
 for (x=0;x<n;x++) {
  l = G->Adj[x] ;
  while (l!=NULL) {
   if (x<=l->but) nb_aretes++ ;
   l = l->suivant ;
  }
 }

 /* on prends toutes les arêtes et on les mets dans L */
 L = malloc (nb_aretes*sizeof(struct __arc)) ;
 i = 0;
 for (x=0;x<n;x++) {
  l = G->Adj[x] ;
  while (l!=NULL) {
   if (x<=l->but) {
     L[i].origine = x;
     L[i].poids = l->poids;
     L[i].but = l->but;
     i++;
   }
   l = l->suivant ;
  }
 }

 for (x=0;x<10;x++) { printf("arete %d : %d --%d--> %d\n", x, L[x].origine, L[x].poids, L[x].but); }
 /* on trie le tableau L */
 qsort(L,nb_aretes,sizeof(struct __arc),cmp__arc);
 for (x=0;x<10;x++) { printf("arete %d : %d --%d--> %d\n", x, L[x].origine, L[x].poids, L[x].but); }


 /* autres initialisations */
 for (x=0;x<n;x++) {
  CC[x] = x ;
  ajoute(x,1,x,&(CCL->Adj[x])) ;
  taille[x] = 1 ;
 }


 /* boucle principale */
 c=0 ; /* indice de l'arete consideree */
 m=0 ; /* nb d'arete deja prises */
 while (m<n-1 && c<nb_aretes) {
  x = L[c].origine;
  y = L[c].but;
  c++;

  if(CC[x] != CC[y]) {
   ajoute(x,1,y,&(T->Adj[x])) ;
   ajoute(y,1,x,&(T->Adj[y])) ;
   m++;

   if (taille[CC[y]] < taille[CC[x]]) { i=CC[y] ; j=CC[x] ;}
   else { j=CC[y] ; i=CC[x] ;}

   while(CCL->Adj[i] != NULL) {
    t=CCL->Adj[i]->but ;
      TMP=CCL->Adj[i] ;
      CCL->Adj[i] = CCL->Adj[i]->suivant ;
      free(TMP) ;
    ajoute(j,1,t,&(CCL->Adj[j])) ;
    CC[t] = CC[j] ;
   }
   taille[y] = taille[x]+taille[y] ;
   taille[x] = 0 ;
  }
 }
 return(T) ;
}



GrapheListe labyrintheRect(int larg, int haut) {
  GrapheListe G = grilleRectPleine(larg,haut) ;
  GrapheListe Labk = kruskal(G);
  return Labk;
}

GrapheListe labyrintheHex(int larg, int haut) {
  GrapheListe G = grilleHexPleine(larg,haut) ;
  GrapheListe Labk = kruskal(G);
  return Labk;
}
