/**********************************************************************
 ***** Pierre Hyvernat, cours info505 "Algorithmes de graphes"    *****
 *****                                                            *****
 ***** Il s'agit des définitions de quelques types de données     *****
 ***** pour le TP en salles machines...                           *****
 ***** Le fichier contient également les prototypes de toutes les *****
 ***** les fonctions définies dans "graphes.c".                   *****
 *********************************************************************/


#ifndef _GRAPHES_H
#define _GRAPHES_H


#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>

/* Une abbréviation pour la liste vide. */
#define NILL NULL

#define NON_VU 0
#define VU 1
#define EXAMINE 2
#define GRIS 0
#define ROUGE 1
#define VERT -1



/* On va utiliser des listes chainées pour les listes d'adjacences.
 * Comme on veut pouvoir faire des graphes avec un poids sur les arêtes,
 * on utilise la définition suivante.
 *
 * Une liste d'adjacence est donc définit par le type "ListeAdj". Si L
 * est une telle liste, on peut accéder aux différents champs avec
 *   - "L->but"    :   but de la première arête
 *   - "L->poids"  :   poids de la première arête
 *   - "L->suivant :   suite de la liste
 *
 * Pour d'autres applications, on aura besoin d'un peu plus :
 *   - "L->origine"
 *   - "L->precedent"
 */


struct __arc {
  int origine ; /* sommet origine de l'arc en question */
  int poids ;   /* poids de l'arête en question */
  int but ;     /* sommet but de l'arc en question */
  struct __arc *suivant ;
  struct __arc *precedent ;
} ;

typedef struct __arc *ListeAdj ;


/* prototype de la fonction "ajoute".
 * Pour faciliter l'insertion, on définit une fonction "ajoute" qui
 * rajoute une arête en tête de liste chainées. Elle prend en arguments
 * une orgine o, un poids "p" et un but "b" et ajoute l'arc correspondant dans
 * la liste.
 * L'argument L doit etre un pointeur pour que l'on puisse modifier la tete de
 * liste. Pour rajouter un arc devant L, il faudra donc utiliser
 * ajoute(o,p,b,&L)
 */
void ajoute (int o, int p, int b, ListeAdj *L) ;



/* La structure de Tile est en fait une liste doublement chainée d'arcs. On
 * pourra donc s'en servir comme d'une file ou d'un pile suivant les fonctions
 * qu'on utilise...
 */

typedef struct __Tile {
  struct __arc *premier ;
  struct __arc *dernier ;
} *Tile ;


/* prototype de la fonction qui initialise une tile vide */
Tile tileVide() ;

/* prototype de la fonction qui insere un arc en tete d'une tile */
void entile(int o, int p, int b, Tile T) ;

/* prototype de la fonction qui enlève le premier élément dans une tile. Si on
 * ne souhaite pas récupérer une des valeur, on peut remplacer son pointeur
 * par NULL.*/
void depile(Tile P, int* o, int* p, int* b) ;

/* prototype de la fonction qui enlève le dernier élément dans une tile. Si on
 * ne souhaite pas récupérer une des valeur, on peut remplacer son pointeur
 * par NULL.**/
void defile(Tile F, int* o, int* p, int* b) ;

/* prototype de la fonction qui teste si une tile est vide */
int tileEstVide(Tile T);


/* Le type des graphes en liste d'adjacences est maintenant simple à
 * définir : un graphes est une structure qui contient un entier "n"
 * (nombre de sommets) et un tableau de listes d'adjacences de longueur
 * "n".
 */
typedef struct __GrapheListe {
 int nb_sommets ;
 ListeAdj *Adj ;
} *GrapheListe ;


/* prototype de la fonction de création d'un graphe en listes
 * d'ajacences.  On définit une fonction qui initialise un graphe vide.
 * Elle prend en argument le nombre de sommets du graphe voulu.
 */
GrapheListe grapheListeVide (int n) ;


/* Le type des graphes en matrices d'adjacences... Par souci de notation, on
 * utilisera un tableau de pointeurs (int **Matr) plutot qu'un tableau
 * unidimensionnel. On pourra donc utiliser G->Matr[i][j] plutot que
 * G->Matr[i*n+j]
 */
typedef struct __GrapheMatr {
 int nb_sommets ;
 int ** Matr ;
} *GrapheMatr ;


/* prototype de la fonction de création d'un graphe en matrice d'adjacence.
 * Une fonction qui initialise une matrice d'adjacence. ATTENTION, cette
 * fonction ne fait que créer la structure. La matrice est complêtement
 * arbitraire...
 */
GrapheMatr grapheMatrVide (int n) ;


/* prototype de la fonction de lecture d'un graphe dans un fichier.
 * Une petite fonction qui permet de lire un graphe dans un fichier. La
 * première ligne du fichier devra comporter un entier "n" qui
 * représentera le nombre de sommets du graphe ; et chaque ligne
 * suivante sera de la forme "a p b" où a est l'origine d'un arc, b le
 * but de l'arc et p le poids de l'arc.
 * L'argument est le nom du fichier, et la valeur de retour un graphe
 * sous forme de liste d'adjacence.
 */
GrapheListe lireGrapheListe(const char* fichier) ;


/* prototype de la fonction de génération aléatoire d'un graphe.  Une
 * petite fonction pour générer un graphe de manière aléatoire.  Cette
 * fonction prend en arguments le nombre de sommets (n), le nombre
 * d'arcs à générer (m), le poids maximum d'un arc (p) et un booléen
 * pour savoir si le graphe doit être orienté ou pas (o).
 */
GrapheListe grapheListeAleatoire(int n, int m, int p, int oriente) ;


/* prototype de la fonction qui initialise une grille rectangulaire pleine */
GrapheListe grilleRectPleine(int larg, int haut) ;

/*prototype d'affichage d'une grille rectangulaire donnée par un graphe.
 * Le dernier argument est un tableau de caractère qui permet d'afficher un
 * caractere dans un sommet donné. (Si ce tableau est NULL, alors on affiche
 * rien du tout.) */
void afficheGrilleRect(GrapheListe G, int larg, int haut, char S[]) ;





/* prototype de la fonction qui initialise une grille hexagonal pleine.
 * La grille aura une hauteur de "haut" et une largeur de "2*lar".
 * Par exemple, l'appel grilleHexPleine(3,5) generera la grille

    1       2       3
  --^--   --^--   --^--
 ___     ___     ___
/  0\___/ 10\___/ 20\___
\___/  5\___/ 15\___/ 25\
/  1\___/ 11\___/ 21\___/
\___/  6\___/ 16\___/ 26\
/  2\___/ 12\___/ 22\___/
\___/  7\___/ 17\___/ 27\
/  3\___/ 13\___/ 23\___/
\___/  8\___/ 18\___/ 28\
/  4\___/ 14\___/ 24\___/
\___/  9\___/ 19\___/ 29\
    \___/   \___/   \___/

 */
GrapheListe grilleHexPleine(int larg, int haut) ;

/*prototype d'affichage d'une grille hexagonale donnée par un graphe
 * Le dernier argument est un tableau de caractère qui permet d'afficher un
 * caractere dans un sommet donné. (Si ce tableau est NULL, alors on affiche
 * rien du tout.) */
void afficheGrilleHex(GrapheListe G, int larg, int haut, char S[]) ;



/* deux fonctions pour generer des labyrinthes sur la grille rectangulaire ou
 * la grille hexagonale. */
GrapheListe labyrintheRect(int larg, int haut);
GrapheListe labyrintheHex(int larg, int haut);


/* L'algorithme de Kruskal pour calculer des arbres couvrants de poids
 * minimal.
 * ATTENTION : si le graphe G n'est pas connexe, on obtient simplement un
 * arbre couvrant minimal de la composante connexe qui contient 0. */
GrapheListe kruskal (GrapheListe G) ;




#endif  /* _GRAPHES_H */
